module Glushkov
-- (Regexp, Regexp', first, headerForm, runInclusion, parseRE, Inclusion, inclusionTA, Inclusion.reverse, aux_prop_inc_refl) 
where

import Control.Monad
import qualified List
import Data.Maybe
import Test.QuickCheck
import qualified Data.Map as Map
import Inclusion
import Char


gl_equal :: Regexp' -> Regexp' -> Maybe Bool
gl_equal r1 r2 = do res1 <- glushkov r1 r2 
                    res2 <- glushkov r2 r1
                    return (res1 && res2)

-- BNF re ::= l | re + re | ( re ) | re re | re*


data Position = Position Int Char
                deriving (Eq, Ord, Show, Read)
  
--instance Eq Position where
--    (==) (Position i l) (Position i' l') = i == i' && l == l'

data Regexp' = Letter' Position | Choice' Regexp' Regexp' | Cat' Regexp' Regexp' | Inter' [Regexp'] | Star' Regexp' | Epsilon'
               deriving (Eq, Ord)

--instance Eq Regexp' where
--    (==) Epsilon' Epsilon' = True
--    (==) (Letter' p1) (Letter' p2) = p1 == p2
--    (==) (Star' r1) (Star' r2) = r1 == r2
--    (==) (Cat' r1 r2) (Cat' r1a r2a) = r1 == r1a && r2 == r2a
--    (==) (Choice' r1 r2) (Choice' r1a r2a) = r1 == r1a && r2 == r2a


instance Arbitrary Regexp' where
    coarbitrary _ a = a
    arbitrary = do 
      r <- arbitrary :: Gen Inclusion.Regexp
      return (mark r)


foldRE' :: a -> (Position -> a) -> (a -> a) -> (a -> a -> a) -> ([a] -> a) -> (a -> a -> a) -> Regexp' -> a
foldRE' eps _ _ _ _ _ Epsilon' = eps
foldRE' _ ch _ _ _ _  (Letter' p) = ch p
foldRE' eps ch str cat int cho (Star' r) = str (foldRE' eps ch str cat int cho r)
foldRE' eps ch str cat int cho (Cat' r1 r2) = cat (foldRE' eps ch str cat int cho r1) (foldRE' eps ch str cat int cho r2)
foldRE' eps ch str cat int cho (Choice' r1 r2) = cho (foldRE' eps ch str cat int cho r1) (foldRE' eps ch str cat int cho r2)
foldRE' eps ch str cat int cho (Inter' rl) = int (foldr (\r -> \s -> (foldRE' eps ch str cat int cho r) : s) [] rl)

instance Show Regexp' where
    show (Letter' (Position i l)) = (show l) ++ "_" ++ (show i)
    show (Choice' r1 r2) = "(" ++ (show r1) ++ "+" ++ (show r2) ++ ")"
    show (Cat' r1 r2) = (show r1) ++ " " ++ (show r2)
    show (Star' r) = "(" ++ (show r) ++ ")*"
    show Epsilon' = "($e)"

-- replaces letters with positions
mark :: Inclusion.Regexp -> Regexp'
mark r = mr where (mr, _) = mark' r 0

data MarkerState a = MarkerState {
      regexp :: a
    , markerCounter :: Integer } deriving (Show)

unmark :: Regexp' -> Inclusion.Regexp
unmark Epsilon' = Epsilon
unmark (Letter' (Position _ l)) = Letter l
unmark (Star' r) = Star (unmark r)
unmark (Cat' r1 r2)   = Cat (unmark r1) (unmark r2)
unmark (Choice' r1 r2)   = Choice (unmark r1) (unmark r2)


mark' :: Inclusion.Regexp -> Int -> (Regexp', Int)
mark' Epsilon i = (Epsilon', i)
mark' (Letter l) i = (Letter' (Position i l), i+1)
mark' (Choice r1 r2) i = 
    let m1 = mark' r1 i
    in let m2 = mark' r2 (snd m1)
    in ((Choice' (fst m1) (fst m2)),(snd m2))
mark' (Cat r1 r2) i = 
    let m1 = mark' r1 i
    in let m2 = mark' r2 (snd m1)
    in ((Cat' (fst m1) (fst m2)),(snd m2))
mark' (Star r1) i = 
    let m1 = mark' r1 i
    in ((Star' (fst m1)),(snd m1))

              
-- First, Last and Follow (McNaughton & Yamada, Bruggemann-Klein & Wood
first :: Regexp' -> [Position]
first Epsilon' = []
first (Letter' p) = [p]
first (Cat' r1 r2) = if (nullable (unmark r1)) then
                        fr1 ++ (first r2)
                    else
                        fr1
    where fr1 = first r1
first (Choice' r1 r2) = (first r1) ++ (first r2)
first (Star' r) = first r

-- auxiliary for last
reverse :: Regexp' -> Regexp'
reverse r = foldRE' 
            Epsilon' 
            (\p -> Letter' p) 
            (\rs -> Star' rs)
            (\r1 -> \r2 -> (Cat' r2 r1))
            (\rl -> Inter' rl)
            (\r1 -> \r2 -> (Choice' r1 r2))
            r

last :: Regexp' -> [Position]
last r = first (Glushkov.reverse r)

symbols :: Regexp' -> [Position]
symbols re = foldRE' [] (\p -> [p]) id (++) (foldr (++) []) (++) re

letters :: Inclusion.Regexp -> [Char]
letters re = foldRE [] (\p -> [p]) id List.union (foldr List.union []) List.union re


-- Follow (From Bruggemann-Klein and Wood
follow :: Regexp' -> Position -> [Position]
follow Epsilon' _ = []
follow (Letter' _) _ = []
follow (Cat' r1 r2) p = if elem p (symbols r1) 
                        then if List.elem p (Glushkov.last r1) 
                             then (follow r1 p) ++ Glushkov.first r2 
                             else follow r1 p
                        else follow r2 p

follow (Choice' r1 r2) p = if elem p (symbols r1) 
                           then follow r1 p 
                           else follow r2 p

follow (Star' r1) p = if List.elem p (Glushkov.last r1) 
                      then remDupl ( follow r1 p ++ Glushkov.first r1  )
                      else follow r1 p

-- Decided whether an re is one-unambiguous
oneUnambiguous :: Regexp' -> Bool
oneUnambiguous re = noDuplicates (map position_char (Glushkov.first re)) && (checkFollow (symbols re) re)

position_char :: Position -> Char
position_char (Position _ c) = c

checkFollow :: [Position] -> Regexp' -> Bool
checkFollow [] _ = True
checkFollow (pos:rs) re = noDuplicates (map position_char (follow re pos)) && checkFollow rs re

remDupl :: Eq a => [a] -> [a]
remDupl l = foldr (\el -> \li -> if elem el li then li else el:li) [] l

createDfa :: Regexp' -> Maybe (Dfa Int)
createDfa re = if oneUnambiguous re 
               then Just (DFA {
                            dfaSigma = alphabet
                          , dfaQi = startStateId
                          , dfaErr = errorStateId 
                          , dfaAcc = acceptStateId
                          , dfaStates = ([(a, STATE {char = c
                                                    , sigma = alphabet
                                                    , delta = foldr (\(Position i ch) m -> Map.insert ch i m) (errorMap alphabet) (follow re (Position a c))
                                                    , final = elem (Position a c) (Glushkov.last re)
                                                    })
                                          | Position a c <- symbols re ]
                                         ++ [(startStateId, STATE {char = ' '
                                                        , sigma=alphabet
                                                        , delta = foldr (\(Position i c) map -> Map.insert c i map) (errorMap alphabet) (Glushkov.first re)
                                                        , final = (Inclusion.nullable (unmark re))
                                                        }
                                             ) 
                                            , (errorStateId, STATE {char = ' '
                                                                   , sigma=alphabet
                                                                   , delta = loopMap errorStateId alphabet
                                                                   , final = False
                                                                   }
                                              )
                                            , (acceptStateId, STATE {char = ' '
                                                                  , sigma=alphabet
                                                                  , delta = loopMap acceptStateId alphabet
                                                                  , final = True
                                                                  }
                                             )
                                            ])
                          })            
               else
                   Nothing
    where alphabet :: [Char] 
          alphabet = letters (unmark re)
          
errorMap :: [Char] -> Map.Map Char Int
errorMap alphabet = loopMap errorStateId alphabet

loopMap :: Eq a => a -> [Char] -> Map.Map Char a
loopMap sId alphabet = foldr (\ch map -> Map.insert ch sId map) Map.empty alphabet


-- memLang :: Dfa Int -> [Char] -> Bool
-- memLang dfa w = memLang' dfa 0 w
-- memLang' :: Dfa Int -> Int -> [Char] -> Bool
-- memLang' dfa q (l:w) = 
-- memLang' (DFA _ _ _ _ _ states) q [] = let DfaState b =  lookup q states 

errorStateId :: Int
errorStateId = -2

startStateId :: Int
startStateId = -1

acceptStateId :: Int
acceptStateId = -3




-- True if there is a word recognized by the Dfa
reachableDFA :: Dfa a -> Bool
reachableDFA (DFA _ qi _ _ l)  =
 fst (reachable_aux qi l [] (lookup qi l))

reachable_aux :: Eq a => a -> [(a, DfaState a)] -> [a] -> Maybe (DfaState a) -> (Bool, [a])
reachable_aux q _ r Nothing = (False, (q:r))
reachable_aux _ _ r (Just (STATE _ _ _ True)) = (True,r)
reachable_aux q d r (Just (STATE _ _ l False)) = callReachable d (q:r) (Prelude.map snd (Map.toList l))

callReachable :: Eq a => [(a, DfaState a)] -> [a] -> [a] -> (Bool, [a])
callReachable _ r [] = (False, r)
callReachable dfa r (s:rs) = if elem s r then callReachable dfa r rs
                             else let (ret, r') = reachable_aux s dfa (s:r) (lookup s dfa) in
                                  if ret then (ret, r')
                                  else callReachable dfa r' rs

-- True if r1 is included in r2
glushkov :: Regexp' -> Regexp' -> Maybe Bool
glushkov r1 r2 = glushkov_aux (createDfa r1) (createDfa r2)

glushkov_aux :: Eq a => (Maybe (Dfa a)) -> (Maybe (Dfa a)) -> (Maybe Bool)
glushkov_aux Nothing _ = Nothing
glushkov_aux _ Nothing = Nothing
glushkov_aux (Just dfa1) (Just dfa2) = (Just (not ( reachableDFA (crossDfa dfa1 (complement (extendDfaSigma dfa2 (dfaSigma dfa1)))))))

data DfaState a = Eq a => 
                  STATE { char :: Char 
                        , sigma :: [Char]
                        , delta :: Map.Map Char a
                        , final :: Bool
                        } 

instance Show (DfaState Int) where
    show state = showDfaState state

showDfaState :: Show a => DfaState a -> String
showDfaState (STATE {char=c, sigma=s, delta=d, final=f}) = "State \'" ++ [c] ++ "\' maps to  " ++ show d ++ " " ++ show f ++ " final, sigma=" ++ show s ++ "\n"

instance Show (DfaState (Int, Int)) where
    show state = showDfaState state

instance Arbitrary (DfaState Int) where
    arbitrary = do
      l <- choose ('a','z') :: Gen Char
      return (STATE {char=l, sigma=[l], delta=Map.empty, final=False})

-- Dfa qi delta final
data Dfa a = Eq a => 
             DFA { dfaSigma :: [Char]
                 , dfaQi :: a 
                 , dfaErr :: a
                 , dfaAcc :: a
                 ,  dfaStates :: [(a, (DfaState a))]
                 }


showDfa :: Show a => (Show (DfaState a) => Dfa a -> String)
showDfa (DFA sig qi err acc l) = "DFA: Sigma:" ++ show sig ++ "\nqI:" ++ show qi ++ "\nerr:" ++ show err ++ "\nacc:" ++ show acc ++ "\n" ++ show l 

instance Show (Dfa Int) where
    show dfa = showDfa dfa

instance Show (Dfa (Int, Int)) where
    show dfa = showDfa dfa

instance Arbitrary (Dfa Int) where
    arbitrary = do
      n_states <- choose (1,1000) :: Gen Int
      return (DFA {dfaSigma = alphabet
                  , dfaQi = 0
                  , dfaErr = -1
                  , dfaAcc = -2
                  , dfaStates = (map (\x -> (x, STATE {char='c',sigma=alphabet, delta= Map.empty, final= False})) (take n_states [1..]))
                  } )
              where alphabet = ['c']

deltaToStates :: Eq a => [a] -> DfaState a -> Bool
deltaToStates _ _ = True

noDuplicates :: Eq a => [a] -> Bool
noDuplicates [] = True
noDuplicates (el:rs) = if elem el rs then False else noDuplicates rs


getStateChar :: Maybe (DfaState a) -> Maybe Char
getStateChar (Just (STATE c _ _ _)) = Just c
getStateChar Nothing = Nothing

getDfaChar :: Eq a => Dfa a -> a -> Maybe Char
getDfaChar dfa a = getStateChar (lookup a (dfaStates dfa))


getState :: Dfa a -> a -> Maybe (DfaState a)
getState (DFA _ _ _ _ states) a = lookup a states

joinDfaState  :: a -> Maybe Char -> b -> Maybe Char -> Maybe (a,b)
joinDfaState  _ Nothing _ _ = Nothing
joinDfaState  _ _ _ Nothing = Nothing
joinDfaState  s1 (Just c1) s2 (Just c2) = if c1 == c2 then Just (s1, s2) else Nothing 


minList :: [Maybe a] -> [a]
minList [] = []
minList (Nothing : r) = minList r
minList ((Just x) : r) = x : minList r

-- Extends the alphabet of a dfa
extendDfaSigma :: Dfa a -> [Char] -> Dfa a
extendDfaSigma (DFA alph qi err acc states) alph2 = (DFA { dfaSigma = newAlph
                                                         , dfaQi = qi 
                                                         , dfaErr = err
                                                         , dfaAcc = acc
                                                         , dfaStates = map (\(sid, STATE {char = ch, sigma = as, delta = d, final = f }) -> 
                                                                                (sid, STATE {char = ch
                                                                                            , sigma = newAlph
                                                                                            , delta = foldr (\ch m -> if elem ch alph 
                                                                                                                      then m 
                                                                                                                      else Map.insert ch (if sid==acc then acc else err) m
                                                                                                            ) d alph2
                                                                                            , final = f}
                                                                                )
                                                                           ) states
                                                         })
    where newAlph :: [Char]
          newAlph = List.union alph alph2



crossState :: a -> DfaState a -> b -> DfaState b -> [Char] -> Maybe ((a,b), DfaState (a,b))
crossState ida (STATE {char=c1,sigma=s1,delta=m1, final=f1}) idb (STATE {char = c2, sigma = _, delta = m2, final = f2}) alph = 
    if c1 == c2 || c1 == ' ' || c2 == ' ' then
        Just ((ida,idb), (STATE {char = c1
                                , sigma = alph
                                , delta  = foldr (\char map -> if elem char alph 
                                                               then pairMap char map (Map.lookup char m1) (Map.lookup char m2)
                                                               else map
                                                 ) Map.empty alph
                                , final = (f1 && f2)}
                         )
             )
                     else
                         Nothing
            where pairMap :: Char -> (Map.Map Char (a,b)) -> (Maybe a) -> (Maybe b) -> (Map.Map Char (a,b))
                  pairMap char map (Just a) (Just b) = Map.insert char (a,b) map

intersection :: Eq a => [a] -> [a] -> [a]
intersection l1 l2 = foldr (\el -> \li -> if elem el l2 then (el:li) else li) [] l1

-- Cross Product of two dfa's
crossDfa :: Dfa a -> Dfa b -> Dfa (a,b)
crossDfa a b = crossDfa' (extendDfaSigma a (dfaSigma b)) (extendDfaSigma b (dfaSigma a))

crossDfa' :: Dfa a -> Dfa b -> Dfa (a,b)
crossDfa' (DFA sigma1 q1 e1 acc1 states1) (DFA _ q2 e2 acc2 states2) = 
    DFA sigma1
            (q1,q2) 
            (e1, e2)
            (acc1, acc2)
            (minList [ crossState id1 s1 id2 s2 sigma1
                           | (id1, s1) <- states1, (id2, s2) <- states2 ])

            
            

                                    
-- dfa recognizing complement language
complement :: Dfa a -> Dfa a
complement (DFA {dfaSigma = s, dfaQi = qi , dfaErr = err , dfaAcc = acc, dfaStates = a}) = 
    DFA {dfaSigma = s, dfaQi=qi, dfaErr=acc, dfaAcc=err
        , dfaStates = [ (sid, STATE {char=c, sigma=alph, delta=d, final=(not f)}) 
                        | (sid, STATE {char= c, sigma=alph, delta= d,final=f}) <- a]}

   
