module InclusionTest
    where
-- QuickCheck tests for the inclusion.hs

import Inclusion
import Glushkov
import qualified Data.Map as Map
import qualified List
import Test.QuickCheck


-- headerForm idempotent
prop_hdf_idemp :: Regexp -> Bool
prop_hdf_idemp re = headerForm (headerForm re) == (headerForm re)


-- runInclusion is reflexive
prop_runInclusion_refl :: Regexp -> Bool
prop_runInclusion_refl re = aux_prop_inc_refl (runInclusion re re)


--reverse id idempotent
prop_reverse_idemp :: Regexp' -> Bool
prop_reverse_idemp re = Glushkov.reverse (Glushkov.reverse re) == re

-- Equality is reflexive

prop_eqre'_refl :: Regexp' -> Bool
prop_eqre'_refl r = r == r

prop_eqre_refl :: Regexp -> Bool
prop_eqre_refl r = r == r


-- OneUnambiguous inclusion correct
prop_oneunamb_inclusion :: Regexp -> Regexp -> Property
prop_oneunamb_inclusion r1 r2 = starNormalForm r1 && starNormalForm r2 ==> 
                                if checkRule ( runInclusion r1 r2 ) == OneAmbiguous
                                then not (oneUnambiguous (mark r1)) || not (oneUnambiguous (mark r2))
                                else True

--oneunmambiuous
prop_onunamb_runincl :: Regexp -> Property
prop_onunamb_runincl re = starNormalForm re ==> if (oneUnambiguous (mark re)) 
                                                then checkRule (runInclusion re re) /= OneAmbiguous
                                                else True
-- parsing is ok
prop_parse_show_idemp :: Regexp -> Bool
prop_parse_show_idemp re = parseRE (show re) == re

-- parsing is ok
prop_show_parse_idemp :: Regexp -> Bool
prop_show_parse_idemp re = parseRE ( show ( (parseRE :: String -> Regexp) (show re))) == re

prop_parse_ok :: Bool
prop_parse_ok = parseRE "aa*+(bc)" == (Choice (Cat (Letter 'a') (Star (Letter 'a'))) (Cat (Letter 'b') (Letter 'c')))

--marking and unmarking are inverses
prop_mark_unmark_ok :: Regexp -> Bool
prop_mark_unmark_ok re = unmark ( mark re) == re


-- symbols mapping
prop_symbols :: Bool
prop_symbols = symbols (mark (Choice (Letter 'a') (Star (Letter 'b')))) == [Position 0 'a', Position 1 'b']

containsAll :: Eq a => [a] -> [a] -> Bool
containsAll [] _ = True
containsAll (el:li) li' = (elem el li') && (containsAll li li')

prop_remDupl :: [Integer] -> Bool
prop_remDupl li = noDuplicates li' && containsAll li li'
                  where li' = remDupl li

-- ANy DFA created from an arbitary regexp should have a word
prop_created_dfa_reachable :: Regexp' -> Bool
prop_created_dfa_reachable re = testDfa (createDfa re) (oneUnambiguous re)
    where testDfa Nothing False = True
          testDfa Nothing True = False
          testDfa (Just dfa) True = reachableDFA dfa
          testDfa (Just _) False = False

-- The cross dfa of the same regexp should have a reachable
prop_crossdfareach :: Regexp' -> Bool
prop_crossdfareach re = reach_aux (createDfa re)
    where reach_aux Nothing = not (oneUnambiguous re)
          reach_aux (Just dfa) = reachableDFA (crossDfa dfa dfa)

-- Check oneUnambiguous
--prop_oneunamb :: Regexp' -> Bool
--prop_oneunamb re = if not (oneUnambiguous re) then 
                       

-- First is unique
prop_first_set :: Regexp' -> Bool
prop_first_set re = noDuplicates (Glushkov.first re) && noDuplicates (Glushkov.last re)



-- DFAs must be valid
checkDfa :: Dfa a -> Bool
checkDfa (DFA alph qi err acc states) = noDuplicates alph
                                        && noDuplicates (Prelude.map fst states)
                                        && elem qi (Prelude.map fst states)
                                        && elem err (Prelude.map fst states)
                                        && elem acc (Prelude.map fst states)
                                        && and (map (checkDfaState alph (Prelude.map fst states)) states)

checkDfaState :: [Char] -> [a] -> (a, DfaState a) -> Bool
checkDfaState alphDfa states (_, STATE {sigma=alphState, delta=d, final=_}) =
    noDuplicates alphState
                     && and (map (\ch -> elem ch alphDfa) alphState)
                     && length alphDfa == length alphState
                     && (and (map (\(a,b) -> elem a alphDfa && elem b states) (Map.toList d))) 
                     && (and (map (\ch -> elem ch (Map.keys d)) alphDfa))


prop_check_created_dfa :: Regexp -> Bool
prop_check_created_dfa re = prop_aux (createDfa (mark re))
    where prop_aux :: Maybe (Dfa a) -> Bool
          prop_aux Nothing = not (oneUnambiguous (mark re))
          prop_aux (Just dfa) = checkDfa dfa




-- An extended dfa must be ok
prop_check_extended_dfa :: Regexp' -> [Char] -> Property
prop_check_extended_dfa re newAlph = oneUnambiguous re ==> prop_aux (createDfa re)
    where prop_aux :: Maybe (Dfa a) -> Bool
          prop_aux Nothing = not (oneUnambiguous re)
          prop_aux (Just dfa) = checkDfa (extendDfaSigma dfa (List.union (dfaSigma dfa) newAlph))

-- inclusion is same
glushkov_inclusion :: Regexp -> Regexp -> Property
glushkov_inclusion r1 r2 = starNormalForm r1 && starNormalForm r2 ==>
                           prop_glushkov_aux r1 r2 (glushkov (mark r1) (mark r2)) 

prop_glushkov_aux :: Regexp -> Regexp -> Maybe Bool -> Bool
prop_glushkov_aux _ _ Nothing = True
prop_glushkov_aux r1 r2 (Just x) = testIncl x (checkRule (runInclusion (headerForm r1) (headerForm r2)))
    where testIncl False NotIncluded = True
          testIncl True NotIncluded = False
          testIncl True Included = True
          testIncl False Included = False
          testIncl _ OneAmbiguous = False


prop_epsilonForm_idemp :: Regexp -> Property
prop_epsilonForm_idemp re = oneUnambiguous (mark re) ==> 
                            gl_equal (mark re) (mark ere) == Just True
                            && isEpsilonForm ere
                               where ere = epsilonForm re

prop_emptyLang :: Regexp -> Bool
prop_emptyLang re = aux (emptyLang re) (letters re)
    where aux True [] = True
          aux False [] = False
          aux False _ = True
          aux True _ = False


-- Tests that runInclusion only fails when the super-expression is one-ambiguous
prop_superoneunambiguous :: Regexp -> Regexp -> Property
prop_superoneunambiguous r1 r2 = starNormalForm r1  && starNormalForm r2 ==>
                                 aux (checkRule (runInclusion (headerForm r1) (headerForm r2)))
    where aux :: Included -> Bool
          aux OneAmbiguous = not (oneUnambiguous (mark r2))
          aux _ = True

incRelExamples :: [(String, String, Included)]
incRelExamples =  [ ("(aaa)*", "(aa)*", NotIncluded)
                  , ("a*b","(a*b)*", Included)
                  , ("a*bb","(a*b)*", Included)
                  , ("a*b","(a*b)+$e", Included)
                  , ("a*","(aa*)+$e", Included)
                  , ("a*b","(aa*b)+$e", NotIncluded)
                  , ("a*b","(aa*b)+b", Included)
                  , ("((aa*b)*c)*", "(aa*+(b+c))*", OneAmbiguous)
                  , ("a* b", "(a+b)*", Included)
                  , ("a*", "a(a(a a* + $e) + $e) + $e", Included)
                  , ("((ab)*c)*", "(ab(ab)*c+c)*", Included)
                  , ("((ab)*c)*", "(ab(ab)*+c)*", OneAmbiguous)
                  , ("(ab(ab)*+c)*", "((ab)*c)*", NotIncluded)
                  , ("(ab(ab)*c+c)*", "((ab)*c)*", Included)
                  , ("a*b*", "(a+b)*", Included)
                  , ("(aaa)*b", "(a(a+b)+b)*", Included)
                  , ("(ab)*", "a*(ba*)*", Included)
                  , ("(ab)*", "a*b*", NotIncluded)
                  , ("((a+b)(c+d))*", "(ac+bd)*", NotIncluded)
                  , ("(ac+bd)*", "((a+b)(c+d))*",  Included)
                  , ("a*b", "(a ((a)*) b )* (b + $e)",  Included)
                  , ("a*+a*b", "a*(b+$e)", Included)
                  , ("(a+b)*", "a(a+b)*+(b(a+b)*+$e)", Included)
                  , ("(a+b+c)*", "(a*b*+b*a*)(c(a+b+c)*)*", OneAmbiguous)
                  , ("(ab)*a", "a(ba)*", Included)
                  , ("a*", "(aa)*", NotIncluded)
                  , ("a*b", "(aa)*b", NotIncluded)
                  , ("a*b", "(a(a+$e))*b", OneAmbiguous)
                  ]

prop_inclusion_predef :: Bool
prop_inclusion_predef = aux incRelExamples
    where aux :: [(String, String, Included)] -> Bool
          aux [] = True
          aux ((r1, r2, b):rs) = inclusionT2 r1 r2 == b && aux rs

print_increl_fault :: [(String, String, Included)] -> String
print_increl_fault [] = ""
print_increl_fault ((r1,r2,res):ls) = if inclusionT2 r1 r2 == res 
                                      then print_increl_fault ls
                                      else show (r1, r2, res) ++ print_increl_fault ls